from aioinject.containers import Container
from aioinject.context import InjectionContext, SyncInjectionContext
from aioinject.decorators import inject
from aioinject.markers import Inject
from aioinject.providers import (
    Callable,
    Factory,
    Object,
    Provider,
    Scoped,
    Singleton,
    Transient,
)


__all__ = [
    "Scoped",
    "Factory",
    "Callable",
    "Container",
    "Inject",
    "InjectionContext",
    "Object",
    "Provider",
    "Singleton",
    "SyncInjectionContext",
    "inject",
    "Transient",
]

__version__ = "0.21.0"



import dataclasses
from typing import Annotated

import pytest

from aioinject import Container, Inject, Scoped


class _A:
    pass


@dataclasses.dataclass
class _B:
    a: Annotated[_A, Inject]


@dataclasses.dataclass
class _C:
    b: Annotated[_B, Inject]


@pytest.fixture
def container() -> Container:
    container = Container()
    container.register(Scoped(_A))
    container.register(Scoped(_B))
    container.register(Scoped(_C))
    return container



import contextlib
from collections.abc import AsyncIterator

import pytest

from aioinject import Scoped, Singleton, providers
from aioinject.containers import Container
from aioinject.context import InjectionContext


class _AbstractService:
    pass


class _ServiceA(_AbstractService):
    pass


class _ServiceB(_AbstractService):
    pass


@pytest.fixture
def container() -> Container:
    return Container()


def test_can_init(container: Container) -> None:
    assert container


def test_can_retrieve_context(container: Container) -> None:
    ctx = container.context()
    assert isinstance(ctx, InjectionContext)


def test_can_register_single(container: Container) -> None:
    provider = providers.Scoped(_ServiceA)
    container.register(provider)

    expected = {_ServiceA: provider}
    assert container.providers == expected


def test_cant_register_multiple_providers_for_same_type(
    container: Container,
) -> None:
    container.register(Scoped(int))

    with pytest.raises(
        ValueError,
        match="^Provider for type <class 'int'> is already registered$",
    ):
        container.register(Scoped(int))


def test_can_retrieve_single_provider(container: Container) -> None:
    int_provider = providers.Scoped(int)
    container.register(int_provider)
    assert container.get_provider(int)


def test_missing_provider() -> None:
    container = Container()
    with pytest.raises(ValueError) as exc_info:  # noqa: PT011
        assert container.get_provider(_ServiceA)

    msg = f"Provider for type {_ServiceA.__qualname__} not found"
    assert str(exc_info.value) == msg


@pytest.mark.anyio
async def test_should_close_singletons() -> None:
    shutdown = False

    @contextlib.asynccontextmanager
    async def dependency() -> AsyncIterator[int]:
        nonlocal shutdown

        yield 42
        shutdown = True

    provider = Singleton(dependency)

    container = Container()
    container.register(provider)
    for _ in range(2):
        async with container.context() as ctx:
            assert await ctx.resolve(int) == 42  # noqa: PLR2004
    assert shutdown is False

    await container.aclose()
    assert shutdown is True



from collections.abc import AsyncIterator
from typing import Any

import httpx
import pytest
from strawberry import Schema
from strawberry.asgi import GraphQL

import aioinject
from aioinject.ext.strawberry import AioInjectExtension
from tests.ext.strawberry.app import StrawberryApp, _Query


@pytest.fixture
def anyio_backend() -> str:
    return "asyncio"


@pytest.fixture
async def app(container: aioinject.Container) -> GraphQL[Any, Any]:
    schema = Schema(
        query=_Query,
        extensions=[AioInjectExtension(container=container)],
    )
    return StrawberryApp(schema=schema)


@pytest.fixture
async def http_client(
    app: GraphQL[Any, Any],
) -> AsyncIterator[httpx.AsyncClient]:
    async with httpx.AsyncClient(app=app, base_url="http://test") as client:
        yield client



import contextlib
from collections.abc import Iterator
from types import TracebackType
from typing import Any, TypeAlias, TypeVar

from typing_extensions import Self

from aioinject._store import SingletonStore
from aioinject.context import InjectionContext, SyncInjectionContext
from aioinject.providers import Provider


_T = TypeVar("_T")
_Providers: TypeAlias = dict[type[_T], Provider[_T]]


class Container:
    def __init__(self) -> None:
        self.providers: _Providers[Any] = {}
        self._singletons = SingletonStore()

    def register(
        self,
        provider: Provider[Any],
    ) -> None:
        if provider.type_ in self.providers:
            msg = f"Provider for type {provider.type_} is already registered"
            raise ValueError(msg)

        self.providers[provider.type_] = provider

    def get_provider(self, type_: type[_T]) -> Provider[_T]:
        try:
            return self.providers[type_]
        except KeyError as exc:
            err_msg = f"Provider for type {type_.__qualname__} not found"
            raise ValueError(err_msg) from exc

    def context(self) -> InjectionContext:
        return InjectionContext(
            container=self,
            singletons=self._singletons,
        )

    def sync_context(self) -> SyncInjectionContext:
        return SyncInjectionContext(
            container=self,
            singletons=self._singletons,
        )

    @contextlib.contextmanager
    def override(
        self,
        provider: Provider[Any],
    ) -> Iterator[None]:
        previous = self.providers.get(provider.type_)
        self.providers[provider.type_] = provider

        yield

        del self.providers[provider.type_]
        if previous is not None:
            self.providers[provider.type_] = previous

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._singletons.__aexit__(exc_type, exc_val, exc_tb)

    async def aclose(self) -> None:
        await self.__aexit__(None, None, None)



from __future__ import annotations

import asyncio
import collections
import contextlib
import enum
import threading
import typing
from collections.abc import AsyncIterator, Iterator
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from types import TracebackType
from typing import TYPE_CHECKING, Any, Literal, TypeVar

from aioinject.providers import DependencyLifetime
from aioinject.utils import enter_context_maybe, enter_sync_context_maybe


if TYPE_CHECKING:
    from typing_extensions import Self

    from aioinject.providers import Provider

T = TypeVar("T")


class NotInCache(enum.Enum):
    sentinel = enum.auto()


class InstanceStore:
    def __init__(self) -> None:
        self._cache: dict[type, Any] = {}
        self._exit_stack = contextlib.AsyncExitStack()
        self._sync_exit_stack = contextlib.ExitStack()

    def get(self, provider: Provider[T]) -> T | Literal[NotInCache.sentinel]:
        return self._cache.get(provider.type_, NotInCache.sentinel)

    def add(self, provider: Provider[T], obj: T) -> None:
        if provider.lifetime is not DependencyLifetime.transient:
            self._cache[provider.type_] = obj

    def lock(
        self,
        provider: Provider[Any],
    ) -> AbstractAsyncContextManager[bool]:
        return contextlib.nullcontext(provider.type_ not in self._cache)

    def sync_lock(
        self,
        provider: Provider[Any],
    ) -> AbstractContextManager[bool]:
        return contextlib.nullcontext(provider.type_ not in self._cache)

    @typing.overload
    async def enter_context(
        self,
        obj: AbstractAsyncContextManager[T] | AbstractContextManager[T],
    ) -> T:
        ...

    @typing.overload
    async def enter_context(self, obj: T) -> T:
        ...

    async def enter_context(
        self,
        obj: AbstractAsyncContextManager[T] | AbstractContextManager[T] | T,
    ) -> T:
        return await enter_context_maybe(obj, self._exit_stack)

    @typing.overload
    def enter_sync_context(self, obj: AbstractContextManager[T]) -> T:
        ...

    @typing.overload
    def enter_sync_context(self, obj: T) -> T:
        ...

    def enter_sync_context(
        self,
        obj: AbstractContextManager[T] | T,
    ) -> T:
        return enter_sync_context_maybe(obj, self._sync_exit_stack)

    async def __aenter__(self) -> Self:
        return self  # pragma: no cover

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    async def aclose(self) -> None:
        await self.__aexit__(None, None, None)

    def __enter__(self) -> Self:
        return self  # pragma: no cover

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._sync_exit_stack.__exit__(exc_type, exc_val, exc_tb)

    def close(self) -> None:
        self.__exit__(None, None, None)


class SingletonStore(InstanceStore):
    def __init__(self) -> None:
        super().__init__()
        self._locks: dict[type, asyncio.Lock] = collections.defaultdict(
            asyncio.Lock,
        )
        self._sync_locks: dict[type, threading.Lock] = collections.defaultdict(
            threading.Lock,
        )

    @contextlib.asynccontextmanager
    async def lock(self, provider: Provider[Any]) -> AsyncIterator[bool]:
        if provider.type_ not in self._cache:
            async with self._locks[provider.type_]:
                yield provider.type_ not in self._cache
                return
        yield False

    @contextlib.contextmanager
    def sync_lock(
        self,
        provider: Provider[Any],
    ) -> Iterator[bool]:
        if provider.type_ not in self._cache:
            with self._sync_locks[provider.type_]:
                yield provider.type_ not in self._cache
                return
        yield False



import contextlib
from collections.abc import AsyncIterator, Iterator
from contextlib import AsyncExitStack, ExitStack

import pytest

from aioinject.utils import enter_context_maybe, enter_sync_context_maybe


pytestmark = pytest.mark.anyio

_NUMBER = 42


@contextlib.asynccontextmanager
async def _async_ctx() -> AsyncIterator[int]:
    yield _NUMBER


@contextlib.contextmanager
def _ctx() -> Iterator[int]:
    yield _NUMBER


async def test_enter_context_maybe() -> None:
    for item in (_async_ctx(), _ctx(), _NUMBER):
        assert await enter_context_maybe(item, AsyncExitStack()) == _NUMBER

    for item in (_ctx(), _NUMBER):
        assert enter_sync_context_maybe(item, ExitStack()) == _NUMBER



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, ParamSpec, TypeVar

from aioinject import decorators, utils


if TYPE_CHECKING:
    from starlette.types import ASGIApp, Receive, Scope, Send

    from aioinject.containers import Container

_T = TypeVar("_T")
_P = ParamSpec("_P")


def inject(function: Callable[_P, _T]) -> Callable[_P, _T]:
    wrapper = decorators.inject(
        function,
        inject_method=decorators.InjectMethod.context,
    )
    return utils.clear_wrapper(wrapper)


class AioInjectMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        container: Container,
    ) -> None:
        self.app = app
        self.container = container

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        async with self.container.context():
            await self.app(scope, receive, send)



from collections.abc import AsyncGenerator, AsyncIterator, Generator, Iterator
from typing import Annotated
from unittest.mock import patch

import pytest

from aioinject import Inject, Provider, providers
from aioinject.providers import Dependency


class _Test:
    pass


@pytest.fixture
def provider() -> Provider[_Test]:
    return providers.Scoped(_Test)


def test_can_provide(provider: Provider[_Test]) -> None:
    instance = provider.provide_sync({})
    assert isinstance(instance, _Test)


def test_would_pass_kwargs_into_impl(provider: Provider[_Test]) -> None:
    with patch.object(provider, "impl") as factory_mock:
        provider.provide_sync({})
        factory_mock.assert_called_once_with()

    kwargs = {"a": 1, "b": 2}
    with patch.object(provider, "impl") as factory_mock:
        provider.provide_sync(kwargs)
        factory_mock.assert_called_once_with(**kwargs)


def test_would_return_factory_result(provider: Provider[_Test]) -> None:
    instance = object()
    with patch.object(provider, "impl") as factory_mock:
        factory_mock.return_value = instance
        assert provider.provide_sync({}) is instance


@pytest.mark.anyio
async def test_provide_async() -> None:
    return_value = 42

    async def factory() -> int:
        return return_value

    provider = providers.Scoped[int](factory)
    assert await provider.provide({}) == return_value


def test_type_hints_on_function() -> None:
    def factory(a: int, b: str) -> None:  # noqa: ARG001
        pass

    provider = providers.Scoped(factory)
    expected = {
        "a": Annotated[int, Inject],
        "b": Annotated[str, Inject],
    }
    assert provider.type_hints == expected


def test_type_hints_on_class() -> None:
    class Test:
        def __init__(self, a: int, b: str) -> None:
            pass

    provider = providers.Scoped(Test)
    expected = {
        "a": Annotated[int, Inject],
        "b": Annotated[str, Inject],
    }
    assert provider.type_hints == expected


def test_annotated_type_hint() -> None:
    def factory(
        a: Annotated[int, Inject()],  # noqa: ARG001
    ) -> None:
        pass

    provider = providers.Scoped(factory)
    assert provider.type_hints == {
        "a": Annotated[int, Inject()],
    }


def test_is_async_on_sync() -> None:
    def factory() -> None:
        pass

    provider = providers.Scoped[None](factory)
    assert not provider.is_async


def test_is_async_on_async() -> None:
    async def factory() -> None:
        pass

    provider = providers.Scoped[None](factory)
    assert provider.is_async


def test_empty_dependencies() -> None:
    def factory() -> None:
        pass

    provider = providers.Scoped(factory)
    assert provider.dependencies == ()


def test_dependencies() -> None:
    def factory(
        a: int,  # noqa: ARG001
        service: Annotated[  # noqa: ARG001
            dict[str, int],
            Inject(),
        ],
        string: Annotated[str, Inject()],  # noqa: ARG001
    ) -> None:
        pass

    provider = providers.Scoped(factory)
    expected = (
        Dependency(
            name="a",
            type_=int,
        ),
        Dependency(
            name="service",
            type_=dict[str, int],
        ),
        Dependency(
            name="string",
            type_=str,
        ),
    )
    assert provider.dependencies == expected


def test_generator_return_types() -> None:
    def iterable() -> Iterator[int]:
        yield 42

    def gen() -> Generator[int, None, None]:
        yield 42

    async def async_iterable() -> AsyncIterator[int]:
        yield 42

    async def async_gen() -> AsyncGenerator[int, None]:
        yield 42

    factories = [iterable, gen, async_iterable, async_gen]
    for factory in factories:
        provider = providers.Scoped(factory)
        assert provider.type_ is int



from __future__ import annotations

import dataclasses
from collections.abc import Sequence
from typing import Annotated, Any

import strawberry
from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket
from strawberry.asgi import GraphQL
from strawberry.dataloader import DataLoader
from strawberry.types import Info

from aioinject import Inject
from aioinject.ext.strawberry import inject
from tests.ext.conftest import NumberService


any_: Any = object()


@strawberry.type
class _Query:
    @strawberry.field
    @inject
    async def hello_world(
        self,
        argument: str,
        provided_value: Annotated[int, Inject],
    ) -> str:
        return f"{argument}-{provided_value}"

    @strawberry.field
    @inject
    def hello_world_sync(
        self,
        argument: str,
        provided_value: Annotated[int, Inject],
    ) -> str:
        return f"{argument}-{provided_value}"

    @strawberry.field
    async def dataloader(self, info: Info[Any, None]) -> Sequence[int]:
        return await info.context.numbers.load_many(list(range(100)))


@inject
async def load_numbers(
    keys: Sequence[int],
    service: Annotated[NumberService, Inject] = any_,
) -> Sequence[int]:
    return [await service.get_number(key) for key in keys]


@dataclasses.dataclass(slots=True, frozen=True)
class Context:
    request: Request | WebSocket
    response: Response

    numbers: DataLoader[int, int] = dataclasses.field(
        default_factory=lambda: DataLoader(load_numbers),
    )


class StrawberryApp(GraphQL[Context, None]):
    async def get_context(
        self,
        request: Request | WebSocket,
        response: Response,
    ) -> Context:
        return Context(
            request=request,
            response=response,
        )



import dataclasses
import inspect
from inspect import Parameter
from typing import Annotated

from aioinject import Inject
from aioinject.ext.strawberry import inject


class _A:
    pass


@dataclasses.dataclass
class _B:
    a: Annotated[Inject, _A]


def dependant(
    arg: int,  # noqa: ARG001
    a: Annotated[_A, Inject],
    b: Annotated[_B, Inject],
) -> tuple[_A, _B]:
    return a, b


def test_should_remove_parameters_from_signature() -> None:
    wrapped = inject(dependant)
    signature = inspect.signature(wrapped)
    assert len(signature.parameters) == 1
    expected = {
        "arg": Parameter(
            name="arg",
            annotation=int,
            kind=Parameter.POSITIONAL_OR_KEYWORD,
        ),
    }
    assert signature.parameters == expected


def test_decorate_method_returning_self() -> None:
    class Class:
        @inject
        def method(self) -> "Class":
            return Class()



from collections.abc import AsyncIterator
from typing import Annotated

import httpx
import pytest
from fastapi import Depends, FastAPI

import aioinject
from aioinject import Inject
from aioinject.ext.fastapi import AioInjectMiddleware, inject


@inject
async def dependency(number: Annotated[int, Inject]) -> AsyncIterator[int]:
    yield number


@pytest.fixture
def app(container: aioinject.Container) -> FastAPI:
    app_ = FastAPI()
    app_.add_middleware(AioInjectMiddleware, container=container)

    @app_.get("/function-route")
    @inject
    async def function_route(
        provided: Annotated[int, Inject],
    ) -> dict[str, str | int]:
        return {"value": provided}

    @app_.get("/depends")
    @inject
    async def route_with_depends(
        number: Annotated[int, Depends(dependency)],
    ) -> dict[str, str | int]:
        return {"value": number}

    return app_


@pytest.fixture
async def http_client(app: FastAPI) -> AsyncIterator[httpx.AsyncClient]:
    async with httpx.AsyncClient(app=app, base_url="http://test") as client:
        yield client



import contextlib
from collections.abc import AsyncGenerator, AsyncIterator, Generator, Iterator
from typing import TypeAlias

import pytest

from aioinject.providers import _guess_return_type


def test_class() -> None:
    class A:
        pass

    assert _guess_return_type(A) is A


def test_function() -> None:
    def factory() -> int:
        return 42

    async def async_factory() -> int:
        return 42

    assert _guess_return_type(factory) is int
    assert _guess_return_type(async_factory) is int  # type: ignore[arg-type]


def test_function_with_no_return_annotation() -> None:
    def factory():  # type: ignore[no-untyped-def] # noqa: ANN202
        pass

    with pytest.raises(ValueError) as exc_info:  # noqa: PT011
        _guess_return_type(factory)

    msg = f"Factory {factory.__qualname__} does not specify return type."
    assert str(exc_info.value) == msg


@pytest.mark.parametrize(
    "return_type",
    [
        Iterator[int],
        Generator[int, None, None],
    ],
)
def test_sync_iterables(return_type: TypeAlias) -> None:
    def iterable() -> return_type:
        yield None

    assert _guess_return_type(iterable) is int
    assert _guess_return_type(contextlib.contextmanager(iterable)) is int  # type: ignore[comparison-overlap]


@pytest.mark.parametrize(
    "return_type",
    [
        AsyncIterator[int],
        AsyncGenerator[int, None],
    ],
)
def test_async_iterables(return_type: TypeAlias) -> None:
    async def iterable() -> return_type:
        yield None

    assert _guess_return_type(iterable) is int
    assert _guess_return_type(contextlib.asynccontextmanager(iterable)) is int  # type: ignore[comparison-overlap]



import contextlib
import inspect
import typing
from collections.abc import Callable, Iterator
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    AsyncExitStack,
    ExitStack,
)
from typing import Any, TypeVar

from aioinject.markers import Inject


_T = TypeVar("_T")
_F = TypeVar("_F", bound=typing.Callable[..., Any])

sentinel = object()


def clear_wrapper(wrapper: _F) -> _F:
    inject_annotations = get_inject_annotations(wrapper)
    signature = inspect.signature(wrapper)
    new_params = tuple(
        p
        for p in signature.parameters.values()
        if p.name not in inject_annotations
    )
    wrapper.__signature__ = signature.replace(  # type: ignore[attr-defined]
        parameters=new_params,
    )
    for name in inject_annotations:
        del wrapper.__annotations__[name]
    return wrapper


def get_inject_annotations(
    function: typing.Callable[..., Any],
) -> dict[str, Any]:
    with remove_annotation(function.__annotations__, "return"):
        return {
            name: annotation
            for name, annotation in typing.get_type_hints(
                function,
                include_extras=True,
            ).items()
            if any(
                isinstance(arg, Inject) or arg is Inject
                for arg in typing.get_args(annotation)
            )
        }


def is_context_manager_function(func: Callable[..., Any]) -> bool:
    while inner := getattr(func, "__wrapped__", None):
        func = inner
    return inspect.isgeneratorfunction(func) or inspect.isasyncgenfunction(
        func,
    )


async def enter_context_maybe(
    resolved: (
        _T | AbstractContextManager[_T] | AbstractAsyncContextManager[_T]
    ),
    stack: AsyncExitStack,
) -> _T:
    if isinstance(resolved, contextlib.ContextDecorator):
        return stack.enter_context(resolved)  # type: ignore[arg-type]

    if isinstance(resolved, contextlib.AsyncContextDecorator):
        return await stack.enter_async_context(
            resolved,  # type: ignore[arg-type]
        )
    return resolved  # type: ignore[return-value]


def enter_sync_context_maybe(
    resolved: (_T | AbstractContextManager[_T]),
    stack: ExitStack,
) -> _T:
    if isinstance(resolved, contextlib.ContextDecorator):
        return stack.enter_context(resolved)  # type: ignore[arg-type]
    return resolved  # type: ignore[return-value]


@contextlib.contextmanager
def remove_annotation(
    annotations: dict[str, Any],
    name: str,
) -> Iterator[None]:
    annotation = annotations.pop(name, sentinel)
    yield
    if annotation is not sentinel:
        annotations[name] = annotation



import contextlib
from collections.abc import AsyncIterator, Generator
from typing import Annotated, Any

import anyio
import pytest

from aioinject import Provider, Scoped, Singleton, providers
from aioinject.containers import Container
from aioinject.markers import Inject


pytestmark = [pytest.mark.anyio]


class _TestError(Exception):
    pass


class _Session:
    pass


class _Repository:
    def __init__(self, session: Annotated[_Session, Inject]) -> None:
        self.session = session


class _Service:
    def __init__(self, repository: Annotated[_Repository, Inject]) -> None:
        self.repository = repository


@pytest.fixture
def container() -> Container:
    container = Container()
    container.register(providers.Scoped(_Session))
    container.register(providers.Scoped(_Repository))
    container.register(providers.Scoped(_Service))
    return container


def test_can_instantiate_context(container: Container) -> None:
    assert container.context()


def test_can_retrieve_service(container: Container) -> None:
    with container.sync_context() as ctx:
        service = ctx.resolve(_Service)
        assert isinstance(service, _Service)
        assert isinstance(service.repository, _Repository)
        assert isinstance(service.repository.session, _Session)


def test_uses_cache(container: Container) -> None:
    with container.sync_context() as ctx:
        service = ctx.resolve(_Service)
        a, b, c = service, service.repository, service.repository.session

        service = ctx.resolve(_Service)
        assert a is service
        assert b is service.repository
        assert c is service.repository.session


def test_does_not_preserve_cache_if_recreated(container: Container) -> None:
    with container.sync_context() as ctx:
        service = ctx.resolve(_Service)

    with container.sync_context() as ctx:
        assert ctx.resolve(_Service) is not service


async def test_provide_async() -> None:
    class Test:
        pass

    container = Container()
    container.register(Scoped(Test))
    async with container.context() as ctx:
        instance = await ctx.resolve(Test)
        assert isinstance(instance, Test)


class _AwaitableCls:
    def __init__(self) -> None:
        self.awaited = False

    def __await__(self) -> Generator[Any, None, None]:
        self.awaited = True
        return anyio.sleep(0).__await__()


async def _async_awaitable() -> _AwaitableCls:
    return _AwaitableCls()


def _sync_awaitable() -> _AwaitableCls:
    return _AwaitableCls()


@pytest.mark.parametrize(
    "provider",
    [
        Scoped(_async_awaitable),  # type: ignore[arg-type]
        Scoped(_sync_awaitable),  # type: ignore[arg-type]
        Singleton(_async_awaitable),  # type: ignore[arg-type]
        Singleton(_sync_awaitable),  # type: ignore[arg-type]
    ],
)
async def test_should_not_execute_awaitable_classes(
    provider: Provider[_AwaitableCls],
) -> None:
    container = Container()
    container.register(provider)

    async with container.context() as ctx:
        resolved = await ctx.resolve(_AwaitableCls)
        assert isinstance(resolved, _AwaitableCls)
        assert not resolved.awaited


async def test_singleton_contextmanager_error() -> None:
    call_number = 0

    @contextlib.asynccontextmanager
    async def raises_error() -> AsyncIterator[int]:
        nonlocal call_number
        call_number += 1
        if call_number == 1:
            raise _TestError
        yield 42

    container = Container()
    container.register(Singleton(raises_error))

    with pytest.raises(_TestError):
        async with container.context() as ctx:
            await ctx.resolve(int)

    async with container.context() as ctx:
        await ctx.resolve(int)



from typing import Annotated, Any

import pytest

from aioinject import Inject, Object


@pytest.mark.anyio
async def test_would_provide_same_object() -> None:
    obj = object()
    provider = Object(object_=obj)

    assert provider.provide_sync({}) is obj
    assert await provider.provide({}) is obj


@pytest.fixture
def dependencies_test_data() -> tuple[Any, ...]:
    class Test:
        def __init__(
            self,
            a: Annotated[int, Inject],
        ) -> None:
            pass

    def test(
        a: Annotated[int, Inject],  # noqa: ARG001
        b: Annotated[Test, Inject],  # noqa: ARG001
    ) -> None:
        pass

    return object(), Test, test


def test_should_have_no_dependencies(
    dependencies_test_data: tuple[Any, ...],
) -> None:
    for obj in dependencies_test_data:
        provider = Object(object_=obj)
        assert not provider.dependencies


def test_should_have_empty_type_hints(
    dependencies_test_data: tuple[Any, ...],
) -> None:
    for obj in dependencies_test_data:
        provider = Object(object_=obj)
        assert not provider.type_hints


def test_should_not_be_async(
    dependencies_test_data: tuple[Any, ...],
) -> None:
    for obj in dependencies_test_data:
        provider = Object(object_=obj)
        assert provider.is_async is False



from __future__ import annotations

import contextlib
from collections.abc import AsyncIterator, Callable
from typing import TYPE_CHECKING, ParamSpec, TypeVar

from litestar import Litestar
from litestar.config.app import AppConfig
from litestar.middleware import MiddlewareProtocol
from litestar.plugins import InitPluginProtocol
from litestar.types import ASGIApp, Receive, Scope, Send

from aioinject import decorators, utils


if TYPE_CHECKING:
    from aioinject.containers import Container

_T = TypeVar("_T")
_P = ParamSpec("_P")

_STATE_KEY = "__aioinject_container__"


def inject(function: Callable[_P, _T]) -> Callable[_P, _T]:
    wrapper = decorators.inject(
        function,
        inject_method=decorators.InjectMethod.context,
    )
    return utils.clear_wrapper(wrapper)


class AioInjectMiddleware(MiddlewareProtocol):
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        app: Litestar = scope["app"]
        container: Container = app.state[_STATE_KEY]
        async with container.context():
            await self.app(scope, receive, send)


class AioInjectPlugin(InitPluginProtocol):
    def __init__(self, container: Container) -> None:
        self.container = container

    @contextlib.asynccontextmanager
    async def _lifespan(
        self,
        _: Litestar,
    ) -> AsyncIterator[None]:
        async with self.container:
            yield

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.state[_STATE_KEY] = self.container
        app_config.middleware.append(AioInjectMiddleware)
        app_config.lifespan.append(self._lifespan)
        return app_config



from collections.abc import Sequence
from typing import Protocol

import aioinject
from aioinject.validation.error import ContainerValidationError


class ContainerValidator(Protocol):
    def __call__(
        self,
        container: aioinject.Container,
    ) -> Sequence[ContainerValidationError]:
        ...



from __future__ import annotations

import collections.abc
import enum
import functools
import inspect
import typing
from collections.abc import Mapping
from dataclasses import dataclass
from inspect import isclass
from typing import (
    Annotated,
    Any,
    ClassVar,
    Generic,
    Protocol,
    TypeAlias,
    TypeVar,
    runtime_checkable,
)

import typing_extensions

from aioinject.markers import Inject
from aioinject.utils import (
    is_context_manager_function,
    remove_annotation,
)


_T = TypeVar("_T")


@dataclass(slots=True, kw_only=True, frozen=True)
class Dependency(Generic[_T]):
    name: str
    type_: type[_T]


def _get_annotation_args(type_hint: Any) -> tuple[type, tuple[Any, ...]]:
    try:
        dep_type, *args = typing.get_args(type_hint)
    except ValueError:
        dep_type, args = type_hint, []
    return dep_type, tuple(args)


def _find_inject_marker_in_annotation_args(
    args: tuple[Any, ...],
) -> Inject | None:
    for arg in args:
        try:
            if issubclass(arg, Inject):
                return Inject()
        except TypeError:
            pass

        if isinstance(arg, Inject):
            return arg
    return None


def collect_dependencies(
    dependant: typing.Callable[..., object] | dict[str, Any],
) -> typing.Iterable[Dependency[object]]:
    if not isinstance(dependant, dict):
        with remove_annotation(dependant.__annotations__, "return"):
            type_hints = typing.get_type_hints(dependant, include_extras=True)
    else:
        type_hints = dependant

    for name, hint in type_hints.items():
        dep_type, args = _get_annotation_args(hint)
        inject_marker = _find_inject_marker_in_annotation_args(args)
        if inject_marker is None:
            continue

        yield Dependency(
            name=name,
            type_=dep_type,
        )


def _get_provider_type_hints(provider: Provider[Any]) -> dict[str, Any]:
    source = provider.impl
    if inspect.isclass(source):
        source = source.__init__

    if isinstance(source, functools.partial):
        return {}

    type_hints = typing.get_type_hints(source, include_extras=True)
    for key, value in type_hints.items():
        _, args = _get_annotation_args(value)
        for arg in args:
            if isinstance(arg, Inject):
                break
        else:
            type_hints[key] = Annotated[value, Inject]

    return type_hints


_GENERATORS = {
    collections.abc.Generator,
    collections.abc.Iterator,
}
_ASYNC_GENERATORS = {
    collections.abc.AsyncGenerator,
    collections.abc.AsyncIterator,
}

_FactoryType: TypeAlias = (
    type[_T]
    | typing.Callable[..., _T]
    | typing.Callable[..., collections.abc.Awaitable[_T]]
    | typing.Callable[..., collections.abc.Coroutine[Any, Any, _T]]
    | typing.Callable[..., collections.abc.Iterator[_T]]
    | typing.Callable[..., collections.abc.AsyncIterator[_T]]
)


def _guess_return_type(factory: _FactoryType[_T]) -> type[_T]:
    if isclass(factory):
        return typing.cast(type[_T], factory)

    type_hints = typing.get_type_hints(factory)
    try:
        return_type = type_hints["return"]
    except KeyError as e:
        msg = f"Factory {factory.__qualname__} does not specify return type."
        raise ValueError(msg) from e

    if origin := typing.get_origin(return_type):
        args = typing.get_args(return_type)

        maybe_wrapped = getattr(  # @functools.wraps
            factory,
            "__wrapped__",
            factory,
        )
        if origin in _ASYNC_GENERATORS and inspect.isasyncgenfunction(
            maybe_wrapped,
        ):
            return args[0]
        if origin in _GENERATORS and inspect.isgeneratorfunction(
            maybe_wrapped,
        ):
            return args[0]

    return return_type


class DependencyLifetime(enum.Enum):
    transient = enum.auto()
    scoped = enum.auto()
    singleton = enum.auto()


@runtime_checkable
class Provider(Protocol[_T]):
    type_: type[_T]
    impl: Any
    lifetime: DependencyLifetime

    async def provide(self, kwargs: Mapping[str, Any]) -> _T:
        ...

    def provide_sync(self, kwargs: Mapping[str, Any]) -> _T:
        ...

    @property
    def type_hints(self) -> dict[str, Any]:
        ...

    @property
    def is_async(self) -> bool:
        ...

    @functools.cached_property
    def dependencies(self) -> tuple[Dependency[object], ...]:
        return tuple(collect_dependencies(self.type_hints))

    @functools.cached_property
    def is_generator(self) -> bool:
        return is_context_manager_function(self.impl)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(type={self.type_}, implementation={self.impl})"


class Scoped(Provider[_T]):
    lifetime = DependencyLifetime.scoped

    def __init__(
        self,
        factory: _FactoryType[_T],
        type_: type[_T] | None = None,
    ) -> None:
        self.type_ = type_ or _guess_return_type(factory)
        self.impl = factory

    def provide_sync(self, kwargs: Mapping[str, Any]) -> _T:
        return self.impl(**kwargs)  # type: ignore[return-value]

    async def provide(self, kwargs: Mapping[str, Any]) -> _T:
        if self.is_async:
            return await self.impl(**kwargs)  # type: ignore[misc]

        return self.provide_sync(kwargs)

    @functools.cached_property
    def type_hints(self) -> dict[str, Any]:
        type_hints = _get_provider_type_hints(self)
        if "return" in type_hints:
            del type_hints["return"]
        return type_hints

    @functools.cached_property
    def is_async(self) -> bool:
        return inspect.iscoroutinefunction(self.impl)


@typing_extensions.deprecated(
    "Callable is deprecated, use aioinject.Scoped instead",
)
class Callable(Scoped[_T]):
    pass


@typing_extensions.deprecated(
    "Factory is deprecated, use aioinject.Scoped instead",
)
class Factory(Scoped[_T]):
    pass


class Singleton(Scoped[_T]):
    lifetime = DependencyLifetime.singleton


class Transient(Scoped[_T]):
    lifetime = DependencyLifetime.transient


class Object(Provider[_T]):
    type_hints: ClassVar[dict[str, Any]] = {}
    is_async = False
    impl: _T
    lifetime = DependencyLifetime.scoped  # It's ok to cache it

    def __init__(
        self,
        object_: _T,
        type_: type[_T] | None = None,
    ) -> None:
        self.type_ = type_ or type(object_)
        self.impl = object_

    def provide_sync(self, kwargs: Mapping[str, Any]) -> _T:  # noqa: ARG002
        return self.impl

    async def provide(self, kwargs: Mapping[str, Any]) -> _T:  # noqa: ARG002
        return self.impl



import dataclasses
from collections.abc import Sequence
from typing import Any


@dataclasses.dataclass(slots=True, frozen=True)
class ContainerValidationError:
    message: str


@dataclasses.dataclass(slots=True, frozen=True)
class DependencyNotFoundError(ContainerValidationError):
    dependency: type[Any]


@dataclasses.dataclass
class ContainerValidationErrorGroup(Exception):  # noqa: N818
    # Exception group could be used instead, but it's added in python 3.11
    errors: Sequence[ContainerValidationError]

    def __str__(self) -> str:
        return repr(self)  # pragma: no cover
